import asyncio
from typing import Awaitable, Callable, Type

import httpx
import pytest
from pydantic import TypeAdapter

from fr24.static import (
    fetch_aircraft_family,
    fetch_airlines,
    fetch_airports,
    fetch_countries,
)
from fr24.types.static import (
    AircraftFamily,
    Airlines,
    Airports,
    Countries,
    StaticData,
)


@pytest.mark.parametrize(
    "fetch_data,static_data_type",
    [
        (fetch_aircraft_family, AircraftFamily),
        (fetch_airlines, Airlines),
        (fetch_airports, Airports),
        (fetch_countries, Countries),
    ],
)
def test_fetch_static_types(
    fetch_data: Callable[[httpx.AsyncClient], Awaitable[StaticData]],
    static_data_type: Type[StaticData],
) -> None:
    async def fetch_data_() -> StaticData:
        async with httpx.AsyncClient() as client:
            return await fetch_data(client)

    data = asyncio.run(fetch_data_())

    ta = TypeAdapter(static_data_type)
    ta.rebuild()
    ta.validate_python(data, extra="forbid", strict=True)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: fr24/proto/_health.proto
# Protobuf Python Version: 6.33.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    33,
    0,
    '',
    'fr24/proto/_health.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x18\x66r24/proto/_health.proto\"\x1c\n\x04Ping\x12\t\n\x01\x61\x18\x01 \x01(\r\x12\t\n\x01\x62\x18\x02 \x01(\r\"\x11\n\x04Pong\x12\t\n\x01\x63\x18\x01 \x01(\rb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'fr24.proto._health_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_PING']._serialized_start=28
  _globals['_PING']._serialized_end=56
  _globals['_PONG']._serialized_start=58
  _globals['_PONG']._serialized_end=75
# @@protoc_insertion_point(module_scope)



#!/usr/bin/env -S uv run --script
# /// script
# requires-python = ">=3.10"
# dependencies = [
#     "fr24[cli]",
# ]
# [tool.uv.sources]
# fr24 = { path = "../", editable = true }
# ///
from __future__ import annotations

import io
import logging
import os
from contextlib import redirect_stdout
from pathlib import Path
from typing import Generator

import click
from typer.main import get_command

from fr24.cli import app

logger = logging.getLogger(__name__)


def process_command(
    command: click.Command, parent_context: list[str]
) -> Generator[str, None, None]:
    command_path = " ".join(parent_context)
    logger.info(f"`{command_path} --help`")
    string_io = io.StringIO()
    with redirect_stdout(string_io):
        command.main(
            ["--help"],
            prog_name=parent_context[0],
            standalone_mode=False,
        )
    help_text = string_io.getvalue().strip().replace("\\[", "[")
    ident = command_path.replace(" ", "_")
    yield f"""--8<-- [start:{ident}]
$ {command_path} --help

{help_text}
--8<-- [end:{ident}]
"""
    if not isinstance(command, click.Group):
        return
    for subcommand_name, subcommand in command.commands.items():
        yield from process_command(
            subcommand, [*parent_context, subcommand_name]
        )


def main() -> None:
    os.environ["COLUMNS"] = "80"
    app.rich_markup_mode = None
    path_docs = Path(__file__).parent.parent / "docs"
    with open(path_docs / "usage" / "cli_output.txt", "w") as f:
        for output in process_command(get_command(app), ["fr24"]):
            f.write(output)


if __name__ == "__main__":
    main()



"""CLI requires Python 3.10+ to work."""
# technically 3.9 works but lack of X | Y is annoying

from __future__ import annotations

import asyncio
import inspect
import logging
import sys
from collections.abc import Sequence
from pathlib import Path
from typing import (
    IO,
    Annotated,
    Any,
    Callable,
    Literal,
    NamedTuple,
    Protocol,
    TypedDict,
    TypeVar,
    get_args,
    get_origin,
    get_type_hints,
)

import click
import typer
from rich.console import Console
from rich.logging import RichHandler
from typing_extensions import ParamSpec, TypeAlias

from . import BBOX_FRANCE_UIR, FR24, BoundingBox, FR24Cache, service
from .cache import PATH_CACHE
from .configuration import FP_CONFIG_FILE, PATH_CONFIG
from .service import LiveFeedPlaybackResult, LiveFeedResult, SupportsWriteTable
from .types import IntFlightId, IntoFlightId, IntoTimestamp
from .types.cache import TabularFileFmt
from .types.grpc import LiveFeedField
from .types.isqx import TimestampS
from .utils import (
    FileExistsBehaviour,
    ParamDetail,
    SphinxParser,
    to_flight_id,
    to_unix_timestamp,
)

stderr = Console(stderr=True)
logging.basicConfig(
    level="INFO",
    format="%(message)s",
    datefmt="[%X]",
    handlers=[RichHandler(console=stderr)],
)
logger = logging.getLogger(__name__)
app = typer.Typer(pretty_exceptions_show_locals=False, no_args_is_help=True)


@app.command()
def dirs() -> None:
    """Shows relevant directories"""
    stderr.print(f"Config: {PATH_CONFIG}")
    stderr.print(f" Cache: {PATH_CACHE}")


@app.command()
def tui() -> None:
    """Starts the TUI"""
    from .tui.tui import main as tui_main

    tui_main()


app_auth = typer.Typer()
app.add_typer(
    app_auth,
    name="auth",
    no_args_is_help=True,
    help="Commands for authentication",
)


@app_auth.command()
def show() -> None:
    """Shows authentication status"""

    ERR_MSG = (
        "[bold yellow]warning[/bold yellow]: not authenticated\n"
        "[bold cyan]help[/bold cyan]: "
        "provide your credentials in environment variables, either:\n"
        "- `fr24_username` and `fr24_password`, or\n"
        "- `fr24_subscription_key` and `fr24_token`\n"
        "[bold cyan]help[/bold cyan]: "
        "alternatively, create a template configuration file "
        f"at `{FP_CONFIG_FILE}` with the command `fr24 auth create`."
    )

    async def show_() -> None:
        async with FR24() as fr24:
            await fr24.login()
            if fr24.http.auth is None:
                stderr.print(ERR_MSG)
            else:
                stderr.print("[bold green]success[/bold green]: authenticated")
                stderr.print(fr24.http.auth)

    asyncio.run(show_())


@app_auth.command()
def create(
    force: Annotated[
        bool,
        typer.Option(help="Overwrite existing configuration file"),
    ] = False,
) -> None:
    """Create a template config file in the user config directory"""
    import shutil

    if FP_CONFIG_FILE.exists() and FP_CONFIG_FILE.is_file() and not force:
        logger.error(
            f"{FP_CONFIG_FILE} already exists, use `--force` to overwrite"
        )
        return

    fp_config_template = (
        Path(__file__).parent.parent.parent / "fr24.example.conf"
    )
    FP_CONFIG_FILE.parent.mkdir(parents=True, exist_ok=True)
    shutil.copy(fp_config_template, FP_CONFIG_FILE)
    stderr.print(
        "[bold green]success[/bold green]: "
        f"created template configuration file at {FP_CONFIG_FILE}"
    )


def get_console(path: Path | IO[bytes] | None) -> Console:
    return Console(stderr=path is not None and not isinstance(path, Path))


#
# generate commands dynamically
#

# in the future, we will allow writing raw data (bytes, proto, json, etc.)
# add `raw` to the `Fmts` and update the defaults accordingly.
Fmts: TypeAlias = tuple[Literal["table"], ...]
FMT_DEFAULT: Fmts = ("table",)

P = ParamSpec("P")
S = TypeVar("S", bound=SupportsWriteTable, covariant=True)


class ServiceCommand(Protocol[P, S]):
    __name__: str

    async def __call__(
        self, fr24: FR24, *args: P.args, **kwargs: P.kwargs
    ) -> S: ...


def register_command(
    app: typer.Typer,
    signature_from: Callable[P, Any],
    *,
    fmts: Fmts = FMT_DEFAULT,
) -> Callable[[ServiceCommand[P, SupportsWriteTable]], Callable[P, None]]:
    def decorator(command_func: ServiceCommand[P, S]) -> Callable[P, None]:
        async def command_body(*args: P.args, **kwargs: P.kwargs) -> None:
            path = kwargs.pop("output")
            format = kwargs.pop("format")
            when_file_exists = kwargs.pop("when_file_exists")

            async with FR24() as fr24:
                await fr24.login()
                result = await command_func(fr24, *args, **kwargs)
                result.write_table(
                    path,  # type: ignore
                    format=format,  # type: ignore
                    when_file_exists=when_file_exists,  # type: ignore
                )

        def cli_command(*args: P.args, **kwargs: P.kwargs) -> None:
            asyncio.run(command_body(*args, **kwargs))

        cli_command.__signature__, cli_command.__doc__ = create_typer_signature(  # type: ignore
            signature_from=signature_from,
            fmts=fmts,
            default_filename=command_func.__name__,
        )
        app.command(name=command_func.__name__.replace("_", "-"))(cli_command)
        return cli_command

    return decorator


class TyperSignature(NamedTuple):
    signature: inspect.Signature
    command_description: str | None


def create_typer_signature(
    signature_from: Any, *, fmts: Fmts, default_filename: str
) -> TyperSignature:
    hints: dict[str, type] = get_type_hints(signature_from, include_extras=True)
    sig = inspect.signature(signature_from)
    command_description: str = ""
    param_docs: dict[str, ParamDetail] = {}
    if docstring := signature_from.__doc__:
        for doc_fragment in SphinxParser(docstring):
            if isinstance(doc_fragment, str):
                command_description += doc_fragment
            elif isinstance(doc_fragment, ParamDetail):
                param_docs[doc_fragment.name] = doc_fragment
    new_params = []
    for param in sig.parameters.values():
        if param.name == "self":
            continue
        detail = param_docs.get(param.name)
        help_text = detail.description if detail else None
        type_hint = hints[param.name]
        param_overrides = get_typer_parameter_override(type_hint)
        kwargs = param_overrides.get("kwargs", {})
        if param.default is inspect.Parameter.empty:
            kwargs["show_default"] = False
        if extra_help := param_overrides.get("extra_help"):
            help_text = (
                extra_help if help_text is None else f"{help_text}{extra_help}"
            )
        new_annotation = Annotated[  # type: ignore
            param_overrides.get("hint") or type_hint,
            typer.Option(help=help_text, **kwargs),
        ]
        new_params.append(param.replace(annotation=new_annotation))
    if fmts:
        default_fmt = "parquet" if "table" in fmts else "json"
        new_params.extend(
            [
                inspect.Parameter(
                    "output",
                    inspect.Parameter.KEYWORD_ONLY,
                    annotation=Annotated[
                        # NOTE: for Annotated[T, typer.Option(parser=P)],
                        # `typer` always tries to build a converter based on `T`
                        # even though a custom parser `P` is provided.
                        Any,  # this disables typer from building a converter
                        typer.Option(
                            "-o",
                            "--output",
                            help=(
                                "Save results to a specific filepath. "
                                "If `-`, results will be printed to stdout. "
                                "If `cache`, results will be saved to the "
                                "default cache."
                            ),
                            parser=FilePathParser(),
                        ),
                    ],
                    default=Path(f"{default_filename}.{default_fmt}"),
                ),
                inspect.Parameter(
                    "format",
                    inspect.Parameter.KEYWORD_ONLY,
                    annotation=Annotated[
                        TabularFileFmt,
                        typer.Option("-f", "--format", help="Output format"),
                    ],
                    default=default_fmt,
                ),
                inspect.Parameter(
                    "when_file_exists",
                    inspect.Parameter.KEYWORD_ONLY,
                    annotation=Annotated[
                        FileExistsBehaviour,
                        typer.Option(
                            help="Action when output file path already exists."
                        ),
                    ],
                    default="backup",
                ),
            ]
        )
    return TyperSignature(
        inspect.Signature(parameters=new_params, return_annotation=None),
        command_description or None,
    )


class BoundingBoxParser(click.ParamType):
    # weirdly, typer reads the `__name__` to show the type in the help message
    __name__ = name = "south,north,west,east"

    def convert(
        self,
        value: Any,
        param: click.Parameter | None,
        ctx: click.Context | None,
    ) -> BoundingBox:
        try:
            parts = tuple(float(p.strip()) for p in str(value).split(","))
        except ValueError as e:
            self.fail(
                f"could not convert {value!r} to BoundingBox: {e}", param, ctx
            )
        if len(parts) != 4:
            self.fail(
                f"expected 4 comma separated values, got {value!r}", param, ctx
            )
        return BoundingBox(*parts)


class TimestampSParser(click.ParamType):
    __name__ = name = "timestamp_s"

    def __init__(self, allow_now: bool) -> None:
        self.allow_now = allow_now

    def convert(
        self,
        value: Any,
        param: click.Parameter | None,
        ctx: click.Context | None,
    ) -> TimestampS[int] | Literal["now"]:
        if (
            isinstance(value, str)
            and value.lower() == "now"
            and not self.allow_now
        ):
            self.fail(
                "this command does not support `now` as a timestamp.",
                param,
                ctx,
            )

        return to_unix_timestamp(value)  # type: ignore


class FlightIdParser(click.ParamType):
    __name__ = name = "flight_id"

    def convert(
        self,
        value: Any,
        param: click.Parameter | None,
        ctx: click.Context | None,
    ) -> IntFlightId:
        try:
            return to_flight_id(value)
        except ValueError as e:
            self.fail(
                f"could not convert {value!r} to FlightId: {e}",
                param,
                ctx,
            )


class FilePathParser(click.ParamType):
    __name__ = name = "filepath|cache|-"

    def convert(
        self,
        value: Any,
        param: click.Parameter | None,
        ctx: click.Context | None,
    ) -> Path | IO[bytes] | FR24Cache:
        if value == "-":
            return sys.stdout.buffer
        if value == "cache":
            return FR24Cache.default()
        path = Path(value).resolve()
        if path.is_dir():
            self.fail(
                f"expected a file path, got a directory: {path!r}",
                param,
                ctx,
            )
        return path


class TyperParamOverride(TypedDict, total=False):
    hint: type | None
    kwargs: dict[str, Any]
    extra_help: str | None


_IntoTimestampArgs = set(get_args(IntoTimestamp))
_IntoFlightIdArgs = set(get_args(IntoFlightId))


def get_typer_parameter_override(type_: type) -> TyperParamOverride:
    """Typer and click do not support `Union` as args. We need to provide custom
    parsers for specific types."""
    origin = get_origin(type_)
    type_args = set(get_args(type_))
    if type_ is BoundingBox:
        example = ", ".join(f"{v:.1f}" for v in BBOX_FRANCE_UIR)
        return {
            "kwargs": {"click_type": BoundingBoxParser()},
            "extra_help": f"Example (france UIR): `{example}`",
        }
    if origin is set and LiveFeedField in type_args:
        return {
            "hint": list[str],
            "kwargs": {"click_type": click.Choice(get_args(LiveFeedField))},
        }
    if _IntoTimestampArgs.issubset(type_args):
        allow_now = Literal["now"] in type_args
        return {
            "hint": int,
            "kwargs": {"click_type": TimestampSParser(allow_now=allow_now)},
        }
    if type_args == _IntoFlightIdArgs:
        return {
            "hint": int,
            "kwargs": {"click_type": FlightIdParser()},
        }
    if origin is Sequence and IntoFlightId in type_args:
        return {
            "hint": list[IntFlightId],
            "kwargs": {"click_type": FlightIdParser()},
        }
    return {}


@register_command(app, signature_from=service.FlightListService.fetch)
async def flight_list(
    fr24: FR24, *args: Any, **kwargs: Any
) -> service.FlightListResult:
    return await fr24.flight_list.fetch(*args, **kwargs)


@register_command(app, signature_from=service.FlightListService.fetch_all)
async def flight_list_all(
    fr24: FR24, *args: Any, **kwargs: Any
) -> service.FlightListResultCollection:
    results = fr24.flight_list.new_result_collection()
    async for result in fr24.flight_list.fetch_all(*args, **kwargs):
        results.append(result)
    return results


@register_command(app, signature_from=service.PlaybackService.fetch)
async def playback(
    fr24: FR24, *args: Any, **kwargs: Any
) -> service.PlaybackResult:
    return await fr24.playback.fetch(*args, **kwargs)


@register_command(app, signature_from=service.LiveFeedService.fetch)
async def live_feed(fr24: FR24, *args: Any, **kwargs: Any) -> LiveFeedResult:
    return await fr24.live_feed.fetch(*args, **kwargs)


@register_command(app, signature_from=service.LiveFeedPlaybackService.fetch)
async def live_feed_playback(
    fr24: FR24, *args: Any, **kwargs: Any
) -> LiveFeedPlaybackResult:
    return await fr24.live_feed_playback.fetch(*args, **kwargs)


# not registering airport_list (unstructured data)
# not registering find (unstructured data)


@register_command(app, signature_from=service.NearestFlightsService.fetch)
async def nearest_flights(
    fr24: FR24, *args: Any, **kwargs: Any
) -> service.NearestFlightsResult:
    return await fr24.nearest_flights.fetch(*args, **kwargs)


@register_command(app, signature_from=service.LiveFlightsStatusService.fetch)
async def live_flights_status(
    fr24: FR24, *args: Any, **kwargs: Any
) -> service.LiveFlightsStatusResult:
    return await fr24.live_flights_status.fetch(*args, **kwargs)


@register_command(app, signature_from=service.FlightDetailsService.fetch)
async def flight_details(
    fr24: FR24, *args: Any, **kwargs: Any
) -> service.FlightDetailsResult:
    return await fr24.flight_details.fetch(*args, **kwargs)


@register_command(app, signature_from=service.TopFlightsService.fetch)
async def top_flights(
    fr24: FR24, *args: Any, **kwargs: Any
) -> service.TopFlightsResult:
    return await fr24.top_flights.fetch(*args, **kwargs)


# not registering follow flight (streaming)


@register_command(app, signature_from=service.PlaybackFlightService.fetch)
async def playback_flight(
    fr24: FR24, *args: Any, **kwargs: Any
) -> service.PlaybackFlightResult:
    return await fr24.playback_flight.fetch(*args, **kwargs)


if __name__ == "__main__":
    app()



import pytest
from pydantic import TypeAdapter

from fr24 import FR24, FR24Cache

FLIGHT_ID = 0x2D81A27


@pytest.mark.anyio
async def test_playback_simple(fr24: FR24) -> None:
    result = await fr24.playback.fetch(flight_id=FLIGHT_ID)
    assert result.to_dict()["result"]["response"]["data"] is not None

    df = result.to_polars()
    assert df.height == 62
    assert df.width == 9

    from fr24.types.json import Playback

    ta = TypeAdapter(Playback)
    ta.rebuild()
    ta.validate_python(result.to_dict(), extra="forbid", strict=True)


@pytest.mark.anyio
async def test_playback_file_ops(fr24: FR24, cache: FR24Cache) -> None:
    """
    check that saving and reopening in a new instance yields the same rows
    """

    result = await fr24.playback.fetch(flight_id=FLIGHT_ID)

    ident = f"{FLIGHT_ID:0x}".upper()
    fp = cache.path / "playback" / f"{ident}.parquet"
    fp.parent.mkdir(parents=True, exist_ok=True)
    fp.unlink(missing_ok=True)
    result.write_table(cache)
    assert fp.exists(), (
        f"{fp} not in {list(f.name for f in fp.parent.glob('*'))}"
    )

    df_local = cache.playback.scan_table(ident).collect()
    assert df_local.equals(result.to_polars())

    # NOTE: making sure that flight metadata is preserved and consistent
    # has been removed.
    # meta_local = datac_local.metadata
    # assert meta_local is not None
    # assert meta_local.get("callsign") == meta.get("callsign")

    fp.unlink()



#!/usr/bin/env -S uv run --script
# /// script
# requires-python = ">=3.10"
# dependencies = [
#     "rich"
# ]
# ///
from __future__ import annotations

import logging
import subprocess
import sys
from pathlib import Path

from rich.logging import RichHandler

logging.basicConfig(
    level="INFO", format="%(message)s", datefmt="[%X]", handlers=[RichHandler()]
)
logger = logging.getLogger(__name__)

PATH_ROOT = Path(__file__).parent.parent.resolve()


def main() -> None:
    """Compile protobuf into `.py` and `.pyi` files."""
    result = subprocess.run(
        ["protoc", "--version"], capture_output=True, text=True, check=True
    )
    protoc_version = result.stdout.strip()
    expected_version = "libprotoc 33.0"
    if protoc_version != expected_version:
        logger.error(
            f"expected `protoc=={expected_version}`, got {protoc_version}"
        )
        sys.exit(1)

    path_mypy_protobuf = PATH_ROOT / ".venv" / "bin" / "protoc-gen-mypy"
    if not path_mypy_protobuf.exists():
        logger.error(f"expected `protoc-gen-mypy` at {path_mypy_protobuf}")
        sys.exit(1)
    path_proto = PATH_ROOT / "src" / "fr24" / "proto"
    proto_files = list(path_proto.glob("*.proto"))

    # NOTE: because protobuf does not support relative imports (https://github.com/protocolbuffers/protobuf/issues/1491)
    # we are forcing the exported package to be `fr24.proto.*`
    # workaround: https://grpc.io/docs/languages/python/basics/#generating-grpc-interfaces-with-custom-package-path
    cmd = [
        "protoc",
        f"--plugin={path_mypy_protobuf}",
        f"-Ifr24/proto={path_proto}",
        "--python_out=src",
        "--mypy_out=readable_stubs:src",
    ] + [str(f) for f in proto_files]
    subprocess.run(cmd, check=True)


if __name__ == "__main__":
    main()



from __future__ import annotations

from datetime import datetime, timezone

from rich.text import Text

from fr24.types.isqx import TimestampS
from fr24.types.json import AircraftInfo
from fr24.types.json import CommonAirport as AirportJSON
from fr24.utils import dataclass_frozen


@dataclass_frozen
class Time:
    timestamp: None | TimestampS[int]

    def __format__(self, __format_spec: str) -> str:
        if self.timestamp is None:
            return ""
        dt = datetime.fromtimestamp(self.timestamp, tz=timezone.utc)
        return format(dt, __format_spec)


def fmt_airport(airport: AirportJSON | None) -> Text:
    if airport is None:
        return Text("")

    code = airport.get("code")
    icao = code.get("icao", "") if code else ""

    city = ""
    if position := airport.get("position"):
        city = position.get("region", {}).get("city", "")

    if not city:
        city = airport.get("name", "")

    if city and icao:
        text = Text(city)
        text.append(" (", style="dim")
        text.append(icao, style="dim")
        text.append(")", style="dim")
        return text
    if city:
        return Text(city)
    if icao:
        text = Text("(", style="dim")
        text.append(icao, style="dim")
        text.append(")", style="dim")
        return text
    return Text("")


def fmt_aircraft(aircraft: AircraftInfo | None) -> Text:
    if aircraft is None:
        return Text("")

    registration = aircraft.get("registration")
    typecode = aircraft.get("model", {}).get("code", "")

    if registration:
        text = Text(registration)
        text.append(" (", style="dim")
        text.append(typecode, style="dim")
        text.append(")", style="dim")
        return text
    return Text(typecode, style="dim")


BLUE = "#5d9ad4"
GREEN = "#67a76d"
ORANGE = "#bd8a44"
RED = "#cf7a78"
GREY = "#959595"


def fmt_status(
    status_text: str,
    *,
    colours: dict[str, str] = {
        "Scheduled": BLUE,
        "Departed": GREEN,
        "Estimated": GREEN,
        "Landed": GREEN,
        "Delayed": ORANGE,
        "Canceled": RED,
        "Diverted": RED,
        "Unknown": GREY,
    },
) -> Text:
    for status, status_color in colours.items():
        if status_text.startswith(status):
            return Text(status_text, style=status_color)
    return Text(status_text, style=GREY)



from __future__ import annotations

from typing import Literal, Union

from typing_extensions import TypedDict

from .isqx import (
    AltitudeFt,
    LatitudeDeg,
    LongitudeDeg,
    TimestampS,
    UtcOffsetS,
)


class Model(TypedDict):
    Name: str
    Code: str


class AircraftFamilyRow(TypedDict):
    description: str
    models: list[Model]


class AircraftFamily(TypedDict):
    version: int
    rows: list[AircraftFamilyRow]


##


class Airline(TypedDict):
    Name: str
    Code: str
    ICAO: str


class Airlines(TypedDict):
    version: int
    rows: list[Airline]


##


class Timezone(TypedDict):
    name: str
    offset: UtcOffsetS[int] | None
    offsetHours: str
    abbr: str
    abbrName: str | None
    isDst: bool


class Airport(TypedDict):
    id: int
    name: str
    iata: str
    icao: str
    city: str
    lat: LatitudeDeg[float]
    lon: LongitudeDeg[float]
    country: str
    alt: AltitudeFt[int] | Literal["-1"]
    size: int
    timezone: Timezone
    countryId: int


class Airports(TypedDict):
    version: str
    rows: list[Airport]


##


class CountryName(TypedDict):
    default: str
    full: str


class CountryCode(TypedDict):
    alpha2: str
    alpha3: str


class Country(TypedDict):
    id: int
    name: CountryName
    code: CountryCode


class Metadata(TypedDict):
    timestamp: TimestampS[int]
    count: int


class Countries(TypedDict):
    metadata: Metadata
    data: list[Country]


##

StaticData = Union[AircraftFamily, Airlines, Airports, Countries]



import logging
import tempfile
from pathlib import Path
from typing import AsyncGenerator

import httpx
import pytest

from fr24 import FR24, FR24Cache


def pytest_configure(config: pytest.Config) -> None:
    try:
        from rich.logging import RichHandler

        handlers = [RichHandler()]
    except ImportError:
        handlers = []
    FORMAT = "%(message)s"
    logging.basicConfig(
        level="INFO", format=FORMAT, datefmt="[%X]", handlers=handlers
    )


@pytest.fixture(scope="session")
def anyio_backend() -> str:
    return "asyncio"


@pytest.fixture(scope="session", autouse=True)
async def client() -> AsyncGenerator[httpx.AsyncClient, None]:
    async with httpx.AsyncClient(http1=False, http2=True) as client:
        yield client


@pytest.fixture(scope="session", autouse=True)
def cache() -> FR24Cache:
    base_dir = Path(tempfile.gettempdir()) / "fr24"
    cache = FR24Cache(base_dir)
    return cache


@pytest.fixture(scope="session", autouse=True)
async def fr24() -> AsyncGenerator[FR24, None]:
    async with FR24() as fr24:
        yield fr24



import pytest

from fr24 import FR24, FR24Cache
from fr24.service import FlightDetailsResult, NearestFlightsResult
from fr24.utils import to_flight_id


@pytest.fixture
async def flight_details_result(
    fr24: FR24,
    nearest_flights_result: NearestFlightsResult,
) -> FlightDetailsResult:
    flight_id = (
        nearest_flights_result.to_proto().flights_list[-1].flight.flightid
    )

    result = await fr24.flight_details.fetch(flight_id=flight_id)
    return result


@pytest.mark.anyio
async def test_flight_details_simple(
    flight_details_result: FlightDetailsResult,
) -> None:
    proto = flight_details_result.to_proto()
    assert len(proto.flight_trail_list) > 10

    df = flight_details_result.to_polars()
    assert df.height == 1
    assert df.width == 32


@pytest.mark.anyio
async def test_flight_details_file_ops(
    flight_details_result: FlightDetailsResult, cache: FR24Cache
) -> None:
    flight_id = f"{to_flight_id(flight_details_result.request.flight_id):0x}"
    timestamp = flight_details_result.timestamp
    fp = (
        cache.path
        / "flight_details"
        / f"{flight_id.upper()}_{timestamp}.parquet"
    )
    fp.parent.mkdir(parents=True, exist_ok=True)
    fp.unlink(missing_ok=True)

    flight_details_result.write_table(cache)
    assert fp.exists(), (
        f"{fp} not in {list(f.name for f in fp.parent.glob('*'))}"
    )

    df_local = cache.flight_details.scan_table(flight_id, timestamp).collect()
    assert df_local.equals(flight_details_result.to_polars())

    fp.unlink()



from __future__ import annotations

import asyncio
from copy import deepcopy
from pathlib import Path
from typing import Callable

import httpx
import orjson
import polars as pl
import pytest
from pydantic import TypeAdapter

from fr24 import FR24, FR24Cache
from fr24.json import (
    FlightListParams,
    PlaybackParams,
    flight_list,
    flight_list_df,
    flight_list_parse,
    get_json_headers,
    playback,
)
from fr24.types.json import FlightList, FlightListItem

REG = "F-HEPK"
FLIGHT = "AF7463"
HEADERS = httpx.Headers(get_json_headers())


def flight_identity(
    flight: FlightListItem,
) -> tuple[str | None, int | None]:
    return (
        flight["identification"]["id"],
        flight["time"]["scheduled"]["departure"],
    )


@pytest.mark.anyio
async def test_ll_flight_list(client: httpx.AsyncClient) -> None:
    list_ = flight_list_parse(
        await flight_list(client, FlightListParams(reg=REG), HEADERS, auth=None)
    ).unwrap()
    df = flight_list_df(list_)
    assert df.shape[0] > 0
    landed = df.filter(pl.col("status").str.starts_with("Landed"))
    if landed.shape[0] == 0:
        return
    result = await asyncio.gather(
        *[
            playback(
                client,
                PlaybackParams(
                    flight_id=flight_id,
                    timestamp=entry["time"]["scheduled"]["arrival"],
                ),
                HEADERS,
                auth=None,
            )
            # the entry below is not None because of `if df is None:`
            for entry in list_["result"]["response"]["data"]  # type: ignore
            if (status := entry["status"]["text"]) is not None
            and status.startswith("Landed")
            and (flight_id := entry["identification"]["id"]) is not None
        ]
    )
    assert len(result) == landed.shape[0]

    ta = TypeAdapter(FlightList)
    ta.rebuild()
    ta.validate_python(list_, extra="forbid", strict=True)


# core


@pytest.mark.anyio
async def test_flight_list_reg(fr24: FR24) -> None:
    with pytest.raises(ValueError):  # missing reg/flight
        _ = await fr24.flight_list.fetch()
    result = await fr24.flight_list.fetch(reg=REG)
    data = result.to_dict()
    flights = data["result"]["response"]["data"]
    assert flights is not None

    df = result.to_polars()
    assert df.height == len(flights)
    assert 0 < df.height <= result.request.limit


@pytest.mark.anyio
async def test_flight_list_flight(fr24: FR24) -> None:
    result = await fr24.flight_list.fetch(flight=FLIGHT)
    data = result.to_dict()
    flights = data["result"]["response"]["data"]
    assert flights is not None

    df = result.to_polars()
    assert df.height == len(flights)
    assert 0 < df.height <= result.request.limit


@pytest.mark.anyio
async def test_flight_list_reg_paginate(fr24: FR24) -> None:
    """
    pagination should never shrink the accumulated unique rows
    """
    results = fr24.flight_list.new_result_collection()
    expected_rows: set[tuple[str | None, int | None]] = set()

    i = 0
    async for result in fr24.flight_list.fetch_all(
        reg=REG, limit=2, delay=5, max_pages=3
    ):
        data = result.to_dict()
        flights = data["result"]["response"]["data"]
        assert flights is not None

        df_new = result.to_polars()
        assert df_new.height > 0
        assert df_new.height == len(flights)

        curr_rows = results.to_polars().height
        assert curr_rows == len(expected_rows)

        expected_rows.update(flight_identity(flight) for flight in flights)
        results.append(result)
        updated_rows = results.to_polars().height
        assert updated_rows == len(expected_rows)
        assert updated_rows >= curr_rows
        assert updated_rows <= curr_rows + df_new.height
        if i > 2:
            assert False, "infinite loop"
        i += 1

    assert results.to_polars().height > 0


@pytest.mark.anyio
async def test_flight_list_reg_concat(fr24: FR24) -> None:
    """
    overlapping pages should be merged by unique flight identity
    """
    result = await fr24.flight_list.fetch(reg=REG, limit=6)
    data = result.to_dict()
    flights = deepcopy(data["result"]["response"]["data"])
    assert flights is not None
    assert len(flights) > 0

    split = max(1, len(flights) // 2)
    left = flights[:split]
    right = flights[max(0, split - 1) :]

    results = fr24.flight_list.new_result_collection()

    data["result"]["response"]["data"] = left
    result.response._content = orjson.dumps(data)
    results.append(deepcopy(result))
    assert results.to_polars().height == len(
        {flight_identity(flight) for flight in left}
    )

    data["result"]["response"]["data"] = right
    result.response._content = orjson.dumps(data)
    results.append(deepcopy(result))
    assert results.to_polars().height == len(
        {flight_identity(flight) for flight in [*left, *right]}
    )


@pytest.mark.anyio
async def test_flight_list_reg_file_ops(fr24: FR24, cache: FR24Cache) -> None:
    """
    check that saving and reopening in a new instance yields the same rows
    test that auto-detect directory and specified directory saving works
    """
    result = await fr24.flight_list.fetch(reg=REG)

    df = result.to_polars()
    curr_rows = df.height

    def test_ok(fp: Path, save: Callable[[], None]) -> None:
        fp.parent.mkdir(parents=True, exist_ok=True)
        fp.unlink(missing_ok=True)
        save()
        assert fp.exists(), (
            f"{fp} not in {list(f.name for f in fp.parent.glob('*'))}"
        )

        df = pl.read_parquet(fp)
        assert df.height == curr_rows
        fp.unlink()

    specific_fp = Path(__file__).parent / "tmp" / "flight_list.parquet"
    test_ok(specific_fp, lambda: result.write_table(specific_fp))
    specific_fp.parent.rmdir()

    test_ok(
        cache.path / "flight_list" / "reg" / f"{REG.upper()}.parquet",
        lambda: result.write_table(cache),
    )



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, Protocol

from platformdirs import user_cache_dir

from .types.cache import (
    flight_details_schema,
    flight_list_schema,
    live_feed_schema,
    live_flights_status_schema,
    nearest_flights_schema,
    playback_flight_schema,
    playback_track_schema,
    top_flights_schema,
)
from .utils import (
    BarePath,
    dataclass_frozen,
    scan_table,
    to_flight_id,
    to_unix_timestamp,
)

if TYPE_CHECKING:
    from typing import Generator

    import polars as pl
    from typing_extensions import Literal

    from .types import IntoFlightId, IntoTimestamp
    from .types.cache import TabularFileFmt
    from .types.isqx import LatitudeDeg, LongitudeDeg


PATH_CACHE = Path(user_cache_dir("fr24"))


class FR24Cache:
    @classmethod
    def default(cls) -> FR24Cache:
        """Create a cache in the
        [default directory](../usage/cli.md#directories)."""
        return cls(PATH_CACHE)

    def __init__(self, path: Path | str) -> None:
        self.path = Path(path)
        flight_list_dir = self.path / "flight_list"

        self.flight_list = FlightListBy(
            reg=FlightListRegCache(Collection(flight_list_dir / "reg")),
            flight=FlightListFlightCache(
                Collection(flight_list_dir / "flight")
            ),
        )
        self.playback = PlaybackCache(Collection(self.path / "playback"))
        self.live_feed = TimestampedCache(
            Collection(self.path / "feed"), schema=live_feed_schema
        )
        self.nearest_flights = NearestFlightsCache(
            Collection(self.path / "nearest_flights")
        )
        self.live_flights_status = TimestampedCache(
            Collection(self.path / "live_flights_status"),
            schema=live_flights_status_schema,
        )
        self.top_flights = TimestampedCache(
            Collection(self.path / "top_flights"), schema=top_flights_schema
        )
        self.flight_details = FlightDetailsCache(
            Collection(self.path / "flight_details"),
            schema=flight_details_schema,
        )
        self.playback_flight = FlightDetailsCache(
            Collection(self.path / "playback_flight"),
            schema=playback_flight_schema,
        )
        # mkdir not necessary: write_table will create them automatically


class CacheLike(Protocol):
    collection: Collection

    def get_path(self, *args: Any, **kwargs: Any) -> BarePath:
        """Get the path to the cached file."""
        ...

    def scan_table(self, *args: Any, **kwargs: Any) -> pl.LazyFrame:
        """Lazily load a file from this collection."""
        ...


@dataclass_frozen
class GlobMixin(CacheLike):
    def glob(self, pattern: str) -> Generator[File, None, None]:
        """Yield all scannable files matching the given pattern."""
        for fp in self.collection.path.glob(pattern):
            yield File(fp)


@dataclass_frozen
class FlightListRegCache(GlobMixin):
    collection: Collection

    def get_path(self, reg: str) -> BarePath:
        return self.collection.new_bare_path(reg.upper())

    def scan_table(
        self, reg: str, *, format: TabularFileFmt = "parquet"
    ) -> pl.LazyFrame:
        """Lazily load a flight list file from this collection.

        :param reg: The aircraft registration number (e.g. `B-HUJ`).
        """
        return scan_table(
            self.get_path(reg), format=format, schema=flight_list_schema
        )


@dataclass_frozen
class FlightListFlightCache(GlobMixin):
    collection: Collection

    def get_path(self, flight: str) -> BarePath:
        return self.collection.new_bare_path(flight.upper())

    def scan_table(
        self, flight: str, *, format: TabularFileFmt = "parquet"
    ) -> pl.LazyFrame:
        """Lazily load a flight list file from this collection.

        :param flight: The flight number (e.g. `CX8747`).
        """
        return scan_table(
            self.get_path(flight), format=format, schema=flight_list_schema
        )


@dataclass_frozen
class FlightListBy:
    reg: FlightListRegCache
    flight: FlightListFlightCache

    def __call__(
        self, kind: Literal["reg", "flight"]
    ) -> FlightListRegCache | FlightListFlightCache:
        return getattr(self, kind)  # type: ignore


@dataclass_frozen
class PlaybackCache(GlobMixin):
    collection: Collection

    def get_path(self, flight_id: IntoFlightId) -> BarePath:
        ident = f"{to_flight_id(flight_id):0x}".upper()
        return self.collection.new_bare_path(ident)

    def scan_table(
        self, flight_id: IntoFlightId, *, format: TabularFileFmt = "parquet"
    ) -> pl.LazyFrame:
        """Lazily load a playback file.

        :param flight_id: The flight ID of the record.
        """
        return scan_table(
            self.get_path(flight_id),
            format=format,
            schema=playback_track_schema,
        )


# not allowing `now` literal for cache because it doesn't make sense
@dataclass_frozen
class TimestampedCache(GlobMixin):
    collection: Collection
    schema: dict[str, pl.DataType] | None

    def get_path(self, timestamp: IntoTimestamp | str) -> BarePath:
        ts = to_unix_timestamp(timestamp)
        if ts is None or ts == "now":
            raise ValueError(f"invalid timestamp for cache: {timestamp}")
        return self.collection.new_bare_path(str(ts))

    def scan_table(
        self,
        timestamp: IntoTimestamp | str,
        *,
        format: TabularFileFmt = "parquet",
    ) -> pl.LazyFrame:
        """Lazily load a timestamped file.

        :param timestamp: A timestamp-like object, see
            [fr24.utils.to_unix_timestamp][]. The `now` literal is not allowed.
        """
        return scan_table(
            self.get_path(timestamp), format=format, schema=self.schema
        )


@dataclass_frozen
class NearestFlightsCache(GlobMixin):
    collection: Collection

    def get_path(
        self,
        lon: LongitudeDeg[float],
        lat: LatitudeDeg[float],
        timestamp: IntoTimestamp | str,
    ) -> BarePath:
        lon6 = int(lon * 1e6)
        lat6 = int(lat * 1e6)
        ts = to_unix_timestamp(timestamp)
        if ts is None or ts == "now":
            raise ValueError(f"invalid timestamp for cache: {timestamp}")
        return self.collection.new_bare_path(f"{lon6}_{lat6}_{ts}")

    def scan_table(
        self,
        lon: LongitudeDeg[float],
        lat: LatitudeDeg[float],
        timestamp: IntoTimestamp | str,
        *,
        format: TabularFileFmt = "parquet",
    ) -> pl.LazyFrame:
        """Lazily load a nearest flights file.

        :param lat: Latitude of the center point.
        :param lon: Longitude of the center point.
        :param timestamp: A timestamp-like object, see
            [fr24.utils.to_unix_timestamp][]. The `now` literal is not allowed.
        """
        return scan_table(
            self.get_path(lon, lat, timestamp),
            format=format,
            schema=nearest_flights_schema,
        )


@dataclass_frozen
class FlightDetailsCache(GlobMixin):
    collection: Collection
    schema: dict[str, pl.DataType] | None

    def get_path(
        self, flight_id: IntoFlightId, timestamp: IntoTimestamp | str
    ) -> BarePath:
        fid = f"{to_flight_id(flight_id):0x}".upper()
        ts = to_unix_timestamp(timestamp)
        if ts is None or ts == "now":
            raise ValueError(f"invalid timestamp for cache: {timestamp}")
        return self.collection.new_bare_path(f"{fid}_{ts}")

    def scan_table(
        self,
        flight_id: IntoFlightId,
        timestamp: IntoTimestamp | str,
        *,
        format: TabularFileFmt = "parquet",
    ) -> pl.LazyFrame:
        """Lazily load a flight details file.

        :param flight_id: The flight ID of the record.
        :param timestamp: A timestamp-like object, see
            [fr24.utils.to_unix_timestamp][]. The `now` literal is not allowed.
        """
        return scan_table(
            self.get_path(flight_id, timestamp),
            format=format,
            schema=self.schema,
        )


@dataclass_frozen
class Collection:
    """A directory containing scannable files."""

    path: Path

    def new_bare_path(self, ident: str) -> BarePath:
        """Returns the bare path (without the file extension) to the file."""
        return BarePath(self.path / ident)


class File(Path):
    @property
    def format(self) -> TabularFileFmt:
        # TODO: perform validation, though it is not critical
        return self.suffix[1:]  # type: ignore

    def scan_table(self) -> pl.LazyFrame:
        """Lazily load this file."""
        return scan_table(self, format=self.format)



"""
Helper functions for the gRPC+protobuf protocol.
For more information, see: https://github.com/grpc/grpc/blob/master/doc/PROTOCOL-HTTP2.md
"""

from __future__ import annotations
from typing import Type, TypeVar
import struct

from google.protobuf.message import Message
from ..utils import Result, Ok, Err
from typing import Union, Protocol
from typing_extensions import runtime_checkable


T_co = TypeVar("T_co", bound=Message, covariant=True)


@runtime_checkable
class SupportsToProto(Protocol[T_co]):
    def to_proto(self) -> T_co:
        """Converts the object into a protobuf message."""


T = TypeVar("T", bound=Message)


# similar to impl Into<Proto>
def to_proto(message_like: SupportsToProto[T] | T) -> T:
    return (
        message_like.to_proto()
        if isinstance(message_like, SupportsToProto)
        else message_like
    )


def encode_message(msg: Message) -> bytes:
    """Encode to a length-prefixed message."""
    msg_bytes = msg.SerializeToString()
    return (
        b"\x00"  # u8, no compression
        + struct.pack(
            ">I", len(msg_bytes)
        )  # u64, length of message, big endian
        + msg_bytes  # binary octet
    )


def parse_data(data: bytes, msg_type: Type[T]) -> Result[T, ProtoError]:
    """Decode a DATA frame (optionally, with Trailers) into a protobuf message."""
    if not data:
        return Err(GrpcError("empty DATA frame", data))
    compressed_flag = data[0]  # 1 byte unsigned int
    if compressed_flag == 1:
        return Err(GrpcError("message is compressed, not implemented", data))
    if compressed_flag != 0:
        try:  # parse trailers
            return Err(GrpcError.from_trailers(data))
        except Exception:
            return Err(
                GrpcError(
                    "unknown message, possibly wrong encoding or non-grpc", data
                )
            )
    data_len = int.from_bytes(data[1:5], byteorder="big")  # message length
    if not data_len:
        return Err(GrpcError("empty message payload", data))
    message = data[5 : 5 + data_len]  # message (in protobuf, binary octet)
    try:
        return Ok(msg_type.FromString(message))
    except Exception as e:
        return Err(ProtoParseError(f"failed to parse message: {e}", data))


class GrpcError(Exception):
    """
    When an application or runtime error occurs during an RPC a
    `Status` and `Status-Message` are delivered in `Trailers`.
    """

    def __init__(
        self,
        message: str,
        raw_data: bytes | None = None,
        *,
        status: int | None = None,
        status_message: bytes | None = None,
        status_details: bytes | None = None,
    ):
        super().__init__(message)
        self.raw_data = raw_data
        """Length-prefixed message"""
        self.status = status
        """1*DIGIT ; 0-9"""
        self.status_message = status_message
        """Percent-Encoded"""
        self.status_details = status_details
        """`google.rpc.Status` proto message"""

    @classmethod
    def from_trailers(cls, data: bytes) -> GrpcError:
        trailers = data[5:]  # skip prefix

        lines = trailers.strip().splitlines()

        status = None
        status_message = None
        status_details = None

        for line in lines:
            if line.startswith(b"grpc-status:"):
                try:
                    status = int(line[12:])
                except ValueError:
                    pass
            elif line.startswith(b"grpc-message:"):
                status_message = line[13:]
            elif line.startswith(b"grpc-status-details-bin:"):
                status_details = line[24:]

        return cls(
            "gRPC errored",
            trailers,
            status=status,
            status_message=status_message,
            status_details=status_details,
        )


class ProtoParseError(Exception):
    def __init__(self, message: str, raw_data: bytes | None = None):
        super().__init__(message)
        self.raw_data = raw_data


ProtoError = Union[GrpcError, ProtoParseError]



"""
Endpoint: https://data-feed.flightradar24.com

Service name: fr24.feed.api.v1.Feed

Methods:

- `LiveFeed`
- `Playback`
- `NearestFlights`
- `LiveFlightsStatus`
- `FollowFlight`
- `TopFlights`
- `LiveTrail`
- ~~`HistoricTrail`~~
- ~~`FetchSearchIndex`~~
- `FlightDetails`
- `PlaybackFlight`

"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, NamedTuple, Sequence, Union

import httpx
from google.protobuf.field_mask_pb2 import FieldMask
from google.protobuf.message import Message

from .proto import (
    ProtoError,
    SupportsToProto,
    encode_message,
    to_proto,
)
from .proto.v1_pb2 import (
    EMSInfo,
    FetchSearchIndexRequest,
    FetchSearchIndexResponse,
    Flight,
    FlightDetailsRequest,
    FlightDetailsResponse,
    FollowedFlight,
    FollowFlightRequest,
    Geolocation,
    HistoricTrailRequest,
    HistoricTrailResponse,
    LiveFeedRequest,
    LiveFeedResponse,
    LiveFlightsStatusRequest,
    LiveFlightsStatusResponse,
    LiveFlightStatus,
    LiveTrailRequest,
    LiveTrailResponse,
    LocationBoundaries,
    NearbyFlight,
    NearestFlightsRequest,
    NearestFlightsResponse,
    PlaybackFlightRequest,
    PlaybackFlightResponse,
    PlaybackRequest,
    PlaybackResponse,
    PositionBuffer,
    RestrictionVisibility,
    TopFlightsRequest,
    TopFlightsResponse,
    TrafficType,
    TrailPoint,
    VisibilitySettings,
)
from .types.isqx import DistanceM, DurationS, LatitudeDeg, LongitudeDeg
from .utils import (
    dataclass_opts,
    get_current_timestamp,
    raise_missing_polars,
    to_flight_id,
    to_unix_timestamp,
)

if TYPE_CHECKING:
    from typing import Annotated, AsyncGenerator, Literal

    import polars as pl
    from google.protobuf.internal.enum_type_wrapper import _V, _EnumTypeWrapper
    from typing_extensions import TypeAlias

    from .types import IntoFlightId, IntoTimestamp
    from .types.cache import (
        EMSRecord,
        FlightDetailsRecord,
        FlightRecord,
        LiveFlightStatusRecord,
        NearbyFlightRecord,
        PlaybackFlightRecord,
        RecentPositionRecord,
        TopFlightRecord,
        TrailPointRecord,
    )
    from .types.grpc import LiveFeedField

#
# helpers
#


def construct_request(
    method_name: str,
    message: Message,
    headers: httpx.Headers,
) -> httpx.Request:
    """Construct the gRPC request with encoded gRPC body."""
    return httpx.Request(
        "POST",
        f"https://data-feed.flightradar24.com/fr24.feed.api.v1.Feed/{method_name}",
        headers=headers,
        content=encode_message(message),
    )


def to_protobuf_enum(
    enum: _V | str | bytes,
    type_wrapper: _EnumTypeWrapper[_V],
) -> _V:
    if isinstance(enum, (str, bytes)):
        return type_wrapper.Value(enum)
    return enum


#
# live feed
#


IntoLiveFeedRequest: TypeAlias = Union[
    SupportsToProto[LiveFeedRequest], LiveFeedRequest
]


class BoundingBox(NamedTuple):
    south: LatitudeDeg[float]
    """Latitude, minimum, degrees"""
    north: LatitudeDeg[float]
    """Latitude, maximum, degrees"""
    west: LongitudeDeg[float]
    """Longitude, minimum, degrees"""
    east: LongitudeDeg[float]
    """Longitude, maximum, degrees"""


@dataclass(**dataclass_opts)
class LiveFeedParams(SupportsToProto[LiveFeedRequest]):
    bounding_box: BoundingBox
    stats: bool = False
    """Whether to include stats in the given area."""
    limit: int = 1500
    """Maximum number of flights (should be set to 1500 for unauthorized users,
    2000 for authorized users).
    """
    maxage: DurationS[int] = 14400
    """Maximum time since last message update, seconds."""
    fields: set[LiveFeedField] = field(
        default_factory=lambda: {"flight", "reg", "route", "type"}
    )
    """Fields to include. For unauthenticated users, a maximum of 4 fields can
    be included.
    When authenticated, `squawk`, `vspeed`, `airspace`, `logo_id` and `age`
    can be included.
    """

    def to_proto(self) -> LiveFeedRequest:
        return LiveFeedRequest(
            bounds=LocationBoundaries(
                north=self.bounding_box.north,
                south=self.bounding_box.south,
                west=self.bounding_box.west,
                east=self.bounding_box.east,
            ),
            settings=VisibilitySettings(
                sources_list=range(10),  # type: ignore
                services_list=range(12),  # type: ignore
                traffic_type=TrafficType.ALL,
                only_restricted=False,
            ),
            field_mask=FieldMask(paths=self.fields),
            highlight_mode=False,
            stats=self.stats,
            limit=self.limit,
            maxage=self.maxage,
            restriction_mode=RestrictionVisibility.NOT_VISIBLE,
        )


async def live_feed(
    client: httpx.AsyncClient,
    message: IntoLiveFeedRequest,
    headers: httpx.Headers,
) -> Annotated[httpx.Response, LiveFeedResponse]:
    request = construct_request("LiveFeed", to_proto(message), headers)
    return await client.send(request)


def live_feed_position_buffer_dict(
    position_buffer: PositionBuffer,
) -> list[RecentPositionRecord]:
    return [
        {
            "delta_lat": pb.delta_lat,
            "delta_lon": pb.delta_lon,
            "delta_ms": pb.delta_ms,
        }
        for pb in position_buffer.recent_positions_list
    ]


def live_feed_flightdata_dict(lfr: Flight) -> FlightRecord:
    """Convert the protobuf message to a dictionary."""
    return {
        "timestamp": lfr.timestamp_ms,
        "flightid": lfr.flightid,
        "latitude": lfr.lat,
        "longitude": lfr.lon,
        "track": lfr.track,
        "altitude": lfr.alt,
        "ground_speed": lfr.speed,
        "on_ground": lfr.on_ground,
        "callsign": lfr.callsign,
        "source": lfr.source,
        "registration": lfr.extra_info.reg,
        "origin": getattr(lfr.extra_info.route, "from"),
        "destination": lfr.extra_info.route.to,
        "typecode": lfr.extra_info.type,
        "eta": lfr.extra_info.schedule.eta,
        "squawk": lfr.extra_info.squawk,
        "vertical_speed": lfr.extra_info.vspeed,
        "position_buffer": live_feed_position_buffer_dict(lfr.position_buffer),
    }


def live_feed_df(
    data: LiveFeedResponse,
) -> pl.DataFrame:
    try:
        import polars as pl
    except ImportError as exc:
        raise_missing_polars(exc)

    from .types.cache import live_feed_schema

    return pl.DataFrame(
        (live_feed_flightdata_dict(lfr) for lfr in data.flights_list),
        schema=live_feed_schema,
    )


#
# live feed playback
#


@dataclass(**dataclass_opts)
class LiveFeedPlaybackParams(SupportsToProto[PlaybackRequest]):
    bounding_box: BoundingBox
    stats: bool = False
    """Whether to include stats in the given area."""
    limit: int = 1500
    """Maximum number of flights (should be set to 1500 for unauthorized users,
    2000 for authorized users).
    """
    maxage: DurationS[int] = 14400
    """Maximum time since last message update, seconds."""
    fields: set[LiveFeedField] = field(
        default_factory=lambda: {"flight", "reg", "route", "type"}
    )
    """Fields to include.
    For unauthenticated users, a maximum of 4 fields can be included.
    When authenticated, `squawk`, `vspeed`, `airspace`, `logo_id` and `age`
    can be included.
    """
    timestamp: IntoTimestamp | Literal["now"] = "now"
    """Start timestamp"""
    duration: DurationS[int] = 7
    """Duration of prefetch, `floor(7.5*(multiplier))` seconds

    For 1x playback, this should be 7 seconds.
    """
    hfreq: int | None = None
    """High frequency mode"""

    def to_proto(self) -> PlaybackRequest:
        timestamp = to_unix_timestamp(self.timestamp)
        if timestamp == "now":
            timestamp = get_current_timestamp() - self.duration
        return PlaybackRequest(
            live_feed_request=LiveFeedParams(
                bounding_box=self.bounding_box,
                stats=self.stats,
                limit=self.limit,
                maxage=self.maxage,
                fields=self.fields,
            ).to_proto(),
            timestamp=timestamp,
            prefetch=timestamp + self.duration,
            hfreq=self.hfreq,
        )


IntoPlaybackRequest: TypeAlias = Union[
    SupportsToProto[PlaybackRequest], PlaybackRequest, LiveFeedPlaybackParams
]


async def live_feed_playback(
    client: httpx.AsyncClient,
    message: IntoPlaybackRequest,
    headers: httpx.Headers,
) -> Annotated[httpx.Response, LiveFeedResponse]:
    request = construct_request("Playback", to_proto(message), headers)
    return await client.send(request)


def live_feed_playback_df(
    data: PlaybackResponse,
) -> pl.DataFrame:
    try:
        import polars as pl
    except ImportError as exc:
        raise_missing_polars(exc)

    from .types.cache import live_feed_schema

    return pl.DataFrame(
        (
            live_feed_flightdata_dict(lfr)
            for lfr in data.live_feed_response.flights_list
        ),
        schema=live_feed_schema,
    )


IntoNearestFlightsRequest: TypeAlias = Union[
    SupportsToProto[NearestFlightsRequest], NearestFlightsRequest
]


@dataclass(**dataclass_opts)
class NearestFlightsParams(SupportsToProto[NearestFlightsRequest]):
    lat: LatitudeDeg[float]
    """Latitude, degrees, -90 to 90"""
    lon: LongitudeDeg[float]
    """Longitude, degrees, -180 to 180"""
    radius: DistanceM[int] = 10000
    """Radius, metres"""
    limit: int = 1500
    """Maximum number of aircraft to return"""

    def to_proto(self) -> NearestFlightsRequest:
        return NearestFlightsRequest(
            location=Geolocation(lat=self.lat, lon=self.lon),
            radius=self.radius,
            limit=self.limit,
        )


async def nearest_flights(
    client: httpx.AsyncClient,
    message: IntoNearestFlightsRequest,
    headers: httpx.Headers,
) -> Annotated[httpx.Response, NearestFlightsResponse]:
    request = construct_request("NearestFlights", to_proto(message), headers)
    return await client.send(request)


IntoLiveFlightsStatusRequest: TypeAlias = Union[
    SupportsToProto[LiveFlightsStatusRequest], LiveFlightsStatusRequest
]


def nearest_flights_nearbyflight_dict(nf: NearbyFlight) -> NearbyFlightRecord:
    return {
        **live_feed_flightdata_dict(nf.flight),
        "distance": nf.distance,
    }


def nearest_flights_df(
    data: NearestFlightsResponse,
) -> pl.DataFrame:
    try:
        import polars as pl
    except ImportError as exc:
        raise_missing_polars(exc)

    from .types.cache import nearest_flights_schema

    return pl.DataFrame(
        (nearest_flights_nearbyflight_dict(nf) for nf in data.flights_list),
        schema=nearest_flights_schema,
    )


@dataclass(**dataclass_opts)
class LiveFlightsStatusParams(SupportsToProto[LiveFlightsStatusRequest]):
    flight_ids: Sequence[IntoFlightId]
    """List of flight IDs to get status for"""

    def to_proto(self) -> LiveFlightsStatusRequest:
        return LiveFlightsStatusRequest(
            flight_ids_list=tuple(to_flight_id(fid) for fid in self.flight_ids)
        )


async def live_flights_status(
    client: httpx.AsyncClient,
    message: IntoLiveFlightsStatusRequest,
    headers: httpx.Headers,
) -> Annotated[httpx.Response, LiveFlightsStatusResponse]:
    request = construct_request("LiveFlightsStatus", to_proto(message), headers)
    return await client.send(request)


def live_flights_status_flightstatusdata_dict(
    flight_status: LiveFlightStatus,
) -> LiveFlightStatusRecord:
    data = flight_status.data

    return {
        "flight_id": flight_status.flight_id,
        "latitude": data.lat,
        "longitude": data.lon,
        "status": data.status,
        "squawk": data.squawk,
    }


def live_flights_status_df(
    data: LiveFlightsStatusResponse,
) -> pl.DataFrame:
    try:
        import polars as pl
    except ImportError as exc:
        raise_missing_polars(exc)

    from .types.cache import live_flights_status_schema

    return pl.DataFrame(
        (
            live_flights_status_flightstatusdata_dict(fs)
            for fs in data.flights_map
        ),
        schema=live_flights_status_schema,
    )


IntoFetchSearchIndexRequest: TypeAlias = Union[
    SupportsToProto[FetchSearchIndexRequest], FetchSearchIndexRequest
]


async def search_index(
    client: httpx.AsyncClient,
    message: IntoFetchSearchIndexRequest,
    headers: httpx.Headers,
) -> Annotated[httpx.Response, FetchSearchIndexResponse]:
    """!!! warning "Unstable API: gateway timeout." """
    request = construct_request("FetchSearchIndex", to_proto(message), headers)
    return await client.send(request)


@dataclass(**dataclass_opts)
class FollowFlightParams(SupportsToProto[FollowFlightRequest]):
    flight_id: IntoFlightId
    """Flight ID to fetch details for.
    Must be live, or the response will contain an empty `DATA` frame error."""
    restriction_mode: RestrictionVisibility.ValueType | str | bytes = (
        RestrictionVisibility.NOT_VISIBLE
    )
    """[FAA LADD](https://www.faa.gov/pilots/ladd) visibility mode."""

    def to_proto(self) -> FollowFlightRequest:
        return FollowFlightRequest(
            flight_id=to_flight_id(self.flight_id),
            restriction_mode=to_protobuf_enum(
                self.restriction_mode, RestrictionVisibility
            ),
        )


IntoFollowFlightRequest: TypeAlias = Union[
    SupportsToProto[FollowFlightRequest], FollowFlightRequest
]


async def follow_flight_stream(
    client: httpx.AsyncClient,
    message: IntoFollowFlightRequest,
    headers: httpx.Headers,
) -> AsyncGenerator[Annotated[bytes, ProtoError]]:
    request = construct_request("FollowFlight", to_proto(message), headers)
    response = await client.send(request, stream=True)
    try:
        async for chunk in response.aiter_bytes():
            yield chunk
    finally:
        await response.aclose()


@dataclass(**dataclass_opts)
class TopFlightsParams(SupportsToProto[TopFlightsRequest]):
    limit: int = 10
    """Maximum number of top flights to return (1-10)"""

    def to_proto(self) -> TopFlightsRequest:
        return TopFlightsRequest(limit=self.limit)


IntoTopFlightsRequest: TypeAlias = Union[
    SupportsToProto[TopFlightsRequest], TopFlightsRequest
]


async def top_flights(
    client: httpx.AsyncClient,
    message: IntoTopFlightsRequest,
    headers: httpx.Headers,
) -> Annotated[httpx.Response, TopFlightsResponse]:
    request = construct_request("TopFlights", to_proto(message), headers)
    return await client.send(request)


def top_flights_dict(ff: FollowedFlight) -> TopFlightRecord:
    return {
        "flight_id": ff.flight_id,
        "live_clicks": ff.live_clicks,
        "total_clicks": ff.total_clicks,
        "flight_number": ff.flight_number,
        "callsign": ff.callsign,
        "squawk": ff.squawk,
        "from_iata": ff.from_iata,
        "from_city": ff.from_city,
        "to_iata": ff.to_iata,
        "to_city": ff.to_city,
        "type": ff.type,
        "full_description": ff.full_description,
    }


def top_flights_df(data: TopFlightsResponse) -> pl.DataFrame:
    try:
        import polars as pl
    except ImportError as exc:
        raise_missing_polars(exc)

    from .types.cache import top_flights_schema

    return pl.DataFrame(
        (top_flights_dict(ff) for ff in data.scoreboard_list),
        schema=top_flights_schema,
    )


IntoLiveTrailRequest: TypeAlias = Union[
    SupportsToProto[LiveTrailRequest], LiveTrailRequest
]


async def live_trail(
    client: httpx.AsyncClient,
    message: IntoLiveTrailRequest,
    headers: httpx.Headers,
) -> Annotated[httpx.Response, LiveTrailResponse]:
    """!!! warning "Unstable API: returns empty `DATA` frame as of Sep 2024"

    Contains empty `DATA` frame error if flight_id is not live"""
    request = construct_request("LiveTrail", to_proto(message), headers)
    return await client.send(request)


IntoHistoricTrailRequest: TypeAlias = Union[
    SupportsToProto[HistoricTrailRequest], HistoricTrailRequest
]


async def historic_trail(
    client: httpx.AsyncClient,
    message: IntoHistoricTrailRequest,
    headers: httpx.Headers,
) -> Annotated[httpx.Response, HistoricTrailResponse]:
    """!!! warning "Unstable API: returns empty `DATA` frame" """
    request = construct_request("HistoricTrail", to_proto(message), headers)
    return await client.send(request)


IntoFlightDetailsRequest: TypeAlias = Union[
    SupportsToProto[FlightDetailsRequest], FlightDetailsRequest
]


@dataclass(**dataclass_opts)
class FlightDetailsParams(SupportsToProto[FlightDetailsRequest]):
    flight_id: IntoFlightId
    """Flight ID to fetch details for.
    Must be live, or the response will contain an empty `DATA` frame error."""
    restriction_mode: RestrictionVisibility.ValueType | str | bytes = (
        RestrictionVisibility.NOT_VISIBLE
    )
    """[FAA LADD](https://www.faa.gov/pilots/ladd) visibility mode."""
    verbose: bool = True
    """Whether to include [fr24.proto.v1_pb2.FlightDetailsResponse.flight_plan]
    and [fr24.proto.v1_pb2.FlightDetailsResponse.aircraft_details] in the
    response."""

    def to_proto(self) -> FlightDetailsRequest:
        return FlightDetailsRequest(
            flight_id=to_flight_id(self.flight_id),
            restriction_mode=to_protobuf_enum(
                self.restriction_mode, RestrictionVisibility
            ),
            verbose=self.verbose,
        )


async def flight_details(
    client: httpx.AsyncClient,
    message: IntoFlightDetailsRequest,
    headers: httpx.Headers,
) -> Annotated[httpx.Response, FlightDetailsResponse]:
    """contains empty `DATA` frame error if flight_id is not live"""
    request = construct_request("FlightDetails", to_proto(message), headers)
    return await client.send(request)


def flight_details_dict(
    response: FlightDetailsResponse,
) -> FlightDetailsRecord:
    aircraft_info = response.aircraft_info
    schedule_info = response.schedule_info
    flight_progress = response.flight_progress
    flight_info = response.flight_info

    return {
        # aircraft info
        "icao_address": aircraft_info.icao_address,
        "reg": aircraft_info.reg,
        "typecode": aircraft_info.type,
        # schedule info
        "flight_number": schedule_info.flight_number,
        "origin_id": schedule_info.origin_id,
        "destination_id": schedule_info.destination_id,
        "diverted_id": schedule_info.diverted_to_id,
        "scheduled_departure": schedule_info.scheduled_departure,
        "scheduled_arrival": schedule_info.scheduled_arrival,
        "actual_departure": schedule_info.actual_departure,
        "actual_arrival": schedule_info.actual_arrival,
        # flight progress
        "traversed_distance": flight_progress.traversed_distance,
        "remaining_distance": flight_progress.remaining_distance,
        "elapsed_time": flight_progress.elapsed_time,
        "remaining_time": flight_progress.remaining_time,
        "eta": flight_progress.eta,
        "great_circle_distance": flight_progress.great_circle_distance,
        "mean_flight_time": flight_progress.mean_flight_time,
        # flight info
        "timestamp_ms": flight_info.timestamp_ms,
        "flightid": flight_info.flightid,
        "latitude": flight_info.lat,
        "longitude": flight_info.lon,
        "track": flight_info.track,
        "altitude": flight_info.alt,
        "ground_speed": flight_info.speed,
        "vertical_speed": flight_info.vspeed,
        "on_ground": flight_info.on_ground,
        "callsign": flight_info.callsign,
        "squawk": flight_info.squawk,
        "ems": ems_dict(flight_info.ems_info),
        # TODO: add flight plan
        "flight_trail_list": [
            trail_point_dict(tp) for tp in response.flight_trail_list
        ],
        "position_buffer": live_feed_position_buffer_dict(
            response.position_buffer
        ),
    }


def trail_point_dict(tp: TrailPoint) -> TrailPointRecord:
    return {
        "timestamp": tp.snapshot_id,
        "latitude": tp.lat,
        "longitude": tp.lon,
        "altitude": tp.altitude,
        "ground_speed": tp.spd,
        "track": tp.heading,
        "vertical_speed": tp.vspd,
        "source": tp.source,
    }


def ems_dict(ems: EMSInfo) -> EMSRecord:
    """Transform Enhanced Mode-S data in the protobuf message into a dictionary.

    This is similar to EMS data in the [JSON API response][fr24.json.playback],
    specifically [fr24.json.playback_track_ems_dict][], which gets converted to
    [fr24.types.cache.PlaybackTrackEMSRecord][]. However, several fields are
    missing:

    - `timestamp`
    - `autopilot`
    - `track`
    - `roll`
    - `precision`
    - `emergency`
    - `tcas_acas`
    - `heading`
    """
    return {
        "ias": ems.ias,
        "tas": ems.tas,
        "mach": ems.mach,
        "mcp": ems.amcp,
        "fms": ems.afms,
        "oat": ems.oat,
        "qnh": ems.qnh,
        "wind_dir": ems.wind_dir,
        "wind_speed": ems.wind_speed,
        "altitude_gps": ems.agps,
        "agpsdiff": ems.agpsdiff,
        "apflags": ems.apflags,
        "rs": ems.rs,
    }


def flight_details_df(
    data: FlightDetailsResponse,
) -> pl.DataFrame:
    try:
        import polars as pl
    except ImportError as exc:
        raise_missing_polars(exc)

    from .types.cache import flight_details_schema

    return pl.DataFrame(
        [flight_details_dict(data)],
        schema=flight_details_schema,
    )


IntoPlaybackFlightRequest: TypeAlias = Union[
    SupportsToProto[PlaybackFlightRequest], PlaybackFlightRequest
]


@dataclass(**dataclass_opts)
class PlaybackFlightParams(SupportsToProto[PlaybackFlightRequest]):
    flight_id: IntoFlightId
    """Flight ID to fetch details for.
    Must not be live, or the response will contain an empty `DATA` frame error.
    """
    timestamp: IntoTimestamp
    """Actual time of departure (ATD) of the historic flight"""

    def to_proto(self) -> PlaybackFlightRequest:
        ts = to_unix_timestamp(self.timestamp)
        assert not isinstance(ts, str)
        return PlaybackFlightRequest(
            flight_id=to_flight_id(self.flight_id),
            timestamp=ts,
        )


async def playback_flight(
    client: httpx.AsyncClient,
    message: IntoPlaybackFlightRequest,
    headers: httpx.Headers,
) -> Annotated[httpx.Response, PlaybackFlightResponse]:
    """contains empty `DATA` frame error if flight_id is live"""
    request = construct_request("PlaybackFlight", to_proto(message), headers)
    return await client.send(request)


def playback_flight_dict(
    response: PlaybackFlightResponse,
) -> PlaybackFlightRecord:
    aircraft_info = response.aircraft_info
    schedule_info = response.schedule_info
    flight_info = response.flight_info

    return {
        # aircraft info
        "icao_address": aircraft_info.icao_address,
        "reg": aircraft_info.reg,
        "typecode": aircraft_info.type,
        # schedule info
        "flight_number": schedule_info.flight_number,
        "origin_id": schedule_info.origin_id,
        "destination_id": schedule_info.destination_id,
        "diverted_id": schedule_info.diverted_to_id,
        "scheduled_departure": schedule_info.scheduled_departure,
        "scheduled_arrival": schedule_info.scheduled_arrival,
        "actual_departure": schedule_info.actual_departure,
        "actual_arrival": schedule_info.actual_arrival,
        # flight info
        "timestamp_ms": flight_info.timestamp_ms,
        "flightid": flight_info.flightid,
        "latitude": flight_info.lat,
        "longitude": flight_info.lon,
        "track": flight_info.track,
        "altitude": flight_info.alt,
        "ground_speed": flight_info.speed,
        "vertical_speed": flight_info.vspeed,
        "on_ground": flight_info.on_ground,
        "callsign": flight_info.callsign,
        "squawk": flight_info.squawk,
        "ems": ems_dict(flight_info.ems_info),
        # flight trail
        "flight_trail_list": [
            trail_point_dict(tp) for tp in response.flight_trail_list
        ],
    }


def playback_flight_df(
    data: PlaybackFlightResponse,
) -> pl.DataFrame:
    try:
        import polars as pl
    except ImportError as exc:
        raise_missing_polars(exc)

    from .types.cache import playback_flight_schema

    return pl.DataFrame(
        [playback_flight_dict(data)],
        schema=playback_flight_schema,
    )



import asyncio
import time

import httpx
import pytest
from google.protobuf.json_format import MessageToDict

from fr24 import BBOX_FRANCE_UIR, BBOXES_WORLD_STATIC, FR24, FR24Cache
from fr24.grpc import (
    BoundingBox,
    LiveFeedParams,
    LiveFeedPlaybackParams,
    live_feed,
    live_feed_playback,
)
from fr24.proto import parse_data
from fr24.proto.headers import get_grpc_headers
from fr24.proto.v1_pb2 import Flight, LiveFeedResponse, PlaybackResponse

HEADERS = httpx.Headers(get_grpc_headers(auth=None))


@pytest.mark.anyio
async def test_ll_live_feed_france(client: httpx.AsyncClient) -> None:
    params = LiveFeedParams(bounding_box=BBOX_FRANCE_UIR)
    response = await live_feed(client, params, HEADERS)
    result = parse_data(response.content, LiveFeedResponse)

    json_output = MessageToDict(
        result.unwrap(), preserving_proto_field_name=True
    )
    assert len(json_output["flights_list"]) > 10


@pytest.mark.anyio
async def test_ll_live_feed_world(client: httpx.AsyncClient) -> None:
    async def get_data(bbox: BoundingBox) -> list[Flight]:
        params = LiveFeedParams(bounding_box=bbox)
        response = await live_feed(client, params, HEADERS)
        result = parse_data(response.content, LiveFeedResponse)
        return list(result.unwrap().flights_list)

    tasks = [get_data(bbox) for bbox in BBOXES_WORLD_STATIC]
    flightss = await asyncio.gather(*tasks)
    assert all(len(f) > 0 for f in flightss)
    assert len([f for flights in flightss for f in flights]) > 100


@pytest.mark.anyio
async def test_ll_live_feed_playback_world(client: httpx.AsyncClient) -> None:
    timestamp = int(time.time() - 86400)

    async def get_data(bbox: BoundingBox) -> list[Flight]:
        params = LiveFeedPlaybackParams(bounding_box=bbox, timestamp=timestamp)
        response = await live_feed_playback(client, params, HEADERS)
        result = parse_data(response.content, PlaybackResponse)
        return list(result.unwrap().live_feed_response.flights_list)

    tasks = [get_data(bbox) for bbox in BBOXES_WORLD_STATIC]
    flightss = await asyncio.gather(*tasks)
    assert all(len(f) > 0 for f in flightss)
    assert len([f for flights in flightss for f in flights]) > 100


# core tests


@pytest.mark.anyio
async def test_live_feed_live_france(fr24: FR24) -> None:
    result = await fr24.live_feed.fetch(BBOX_FRANCE_UIR)
    len_proto = len(result.to_proto().flights_list)
    assert len_proto > 10

    df = result.to_polars()
    assert df.height == len_proto


@pytest.mark.anyio
async def test_live_feed_playback_france(fr24: FR24) -> None:
    yesterday = int(time.time() - 86400)
    result = await fr24.live_feed_playback.fetch(
        BBOX_FRANCE_UIR, timestamp=yesterday
    )
    assert result.to_polars().height > 10


@pytest.mark.anyio
async def test_live_feed_file_ops(fr24: FR24, cache: FR24Cache) -> None:
    """ensure context persists after serialisation to parquet"""
    result = await fr24.live_feed.fetch(BBOX_FRANCE_UIR)
    result.write_table(cache)

    df_local = cache.live_feed.scan_table(result.timestamp).collect()
    assert df_local.equals(result.to_polars())
    # NOTE: metadata serialisation is not implemented



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

from google.protobuf import descriptor as descriptor
from google.protobuf import message as message
import builtins as builtins
import sys
import typing as typing

if sys.version_info >= (3, 10):
    from typing import TypeAlias as _TypeAlias
else:
    from typing_extensions import TypeAlias as _TypeAlias

DESCRIPTOR: descriptor.FileDescriptor

@typing.final
class Ping(message.Message):
    DESCRIPTOR: descriptor.Descriptor

    A_FIELD_NUMBER: builtins.int
    B_FIELD_NUMBER: builtins.int
    a: builtins.int
    b: builtins.int
    def __init__(
        self,
        *,
        a: builtins.int = ...,
        b: builtins.int = ...,
    ) -> None: ...
    _ClearFieldArgType: _TypeAlias = typing.Literal["a", b"a", "b", b"b"]  # noqa: Y015
    def ClearField(self, field_name: _ClearFieldArgType) -> None: ...

@typing.final
class Pong(message.Message):
    DESCRIPTOR: descriptor.Descriptor

    C_FIELD_NUMBER: builtins.int
    c: builtins.int
    def __init__(
        self,
        *,
        c: builtins.int = ...,
    ) -> None: ...
    _ClearFieldArgType: _TypeAlias = typing.Literal["c", b"c"]  # noqa: Y015
    def ClearField(self, field_name: _ClearFieldArgType) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: fr24/proto/v1.proto
# Protobuf Python Version: 6.33.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    33,
    0,
    '',
    'fr24/proto/v1.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from fr24.proto import _common_pb2 as fr24_dot_proto_dot___common__pb2
from fr24.proto import _live_feed_pb2 as fr24_dot_proto_dot___live__feed__pb2
from fr24.proto import _health_pb2 as fr24_dot_proto_dot___health__pb2
from fr24.proto import _nearest_flights_pb2 as fr24_dot_proto_dot___nearest__flights__pb2
from fr24.proto import _live_flight_status_pb2 as fr24_dot_proto_dot___live__flight__status__pb2
from fr24.proto import _fetch_search_index_pb2 as fr24_dot_proto_dot___fetch__search__index__pb2
from fr24.proto import _follow_flight_pb2 as fr24_dot_proto_dot___follow__flight__pb2
from fr24.proto import _top_flights_pb2 as fr24_dot_proto_dot___top__flights__pb2
from fr24.proto import _live_trail_pb2 as fr24_dot_proto_dot___live__trail__pb2
from fr24.proto import _historic_trail_pb2 as fr24_dot_proto_dot___historic__trail__pb2
from fr24.proto import _flight_details_pb2 as fr24_dot_proto_dot___flight__details__pb2
from fr24.proto import _playback_flight_pb2 as fr24_dot_proto_dot___playback__flight__pb2

from fr24.proto._common_pb2 import *
from fr24.proto._live_feed_pb2 import *
from fr24.proto._health_pb2 import *
from fr24.proto._nearest_flights_pb2 import *
from fr24.proto._live_flight_status_pb2 import *
from fr24.proto._fetch_search_index_pb2 import *
from fr24.proto._follow_flight_pb2 import *
from fr24.proto._top_flights_pb2 import *
from fr24.proto._live_trail_pb2 import *
from fr24.proto._historic_trail_pb2 import *
from fr24.proto._flight_details_pb2 import *
from fr24.proto._playback_flight_pb2 import *

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x13\x66r24/proto/v1.proto\x12\x10\x66r24.feed.api.v1\x1a\x18\x66r24/proto/_common.proto\x1a\x1b\x66r24/proto/_live_feed.proto\x1a\x18\x66r24/proto/_health.proto\x1a!fr24/proto/_nearest_flights.proto\x1a$fr24/proto/_live_flight_status.proto\x1a$fr24/proto/_fetch_search_index.proto\x1a\x1f\x66r24/proto/_follow_flight.proto\x1a\x1d\x66r24/proto/_top_flights.proto\x1a\x1c\x66r24/proto/_live_trail.proto\x1a fr24/proto/_historic_trail.proto\x1a fr24/proto/_flight_details.proto\x1a!fr24/proto/_playback_flight.proto2\xf4\x05\n\x04\x46\x65\x65\x64\x12\x14\n\x04\x45\x63ho\x12\x05.Ping\x1a\x05.Pong\x12/\n\tCountDown\x12\x11._common.Duration\x1a\r._common.Tick0\x01\x12/\n\x08LiveFeed\x12\x10.LiveFeedRequest\x1a\x11.LiveFeedResponse\x12/\n\x08Playback\x12\x10.PlaybackRequest\x1a\x11.PlaybackResponse\x12\x41\n\x0eNearestFlights\x12\x16.NearestFlightsRequest\x1a\x17.NearestFlightsResponse\x12J\n\x11LiveFlightsStatus\x12\x19.LiveFlightsStatusRequest\x1a\x1a.LiveFlightsStatusResponse\x12G\n\x10\x46\x65tchSearchIndex\x12\x18.FetchSearchIndexRequest\x1a\x19.FetchSearchIndexResponse\x12=\n\x0c\x46ollowFlight\x12\x14.FollowFlightRequest\x1a\x15.FollowFlightResponse0\x01\x12\x35\n\nTopFlights\x12\x12.TopFlightsRequest\x1a\x13.TopFlightsResponse\x12\x32\n\tLiveTrail\x12\x11.LiveTrailRequest\x1a\x12.LiveTrailResponse\x12>\n\rHistoricTrail\x12\x15.HistoricTrailRequest\x1a\x16.HistoricTrailResponse\x12>\n\rFlightDetails\x12\x15.FlightDetailsRequest\x1a\x16.FlightDetailsResponse\x12\x41\n\x0ePlaybackFlight\x12\x16.PlaybackFlightRequest\x1a\x17.PlaybackFlightResponseB\x02H\x01P\x00P\x01P\x02P\x03P\x04P\x05P\x06P\x07P\x08P\tP\nP\x0b\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'fr24.proto.v1_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'H\001'
  _globals['_FEED']._serialized_start=431
  _globals['_FEED']._serialized_end=1187
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

from typing import TypeVar

import isqx
import isqx.aerospace
import isqx.usc
from typing_extensions import Annotated, TypeAlias

_T = TypeVar("_T")


TimestampS: TypeAlias = Annotated[_T, isqx.TIME[isqx.S]]
TimestampMs: TypeAlias = Annotated[_T, isqx.TIME[isqx.MILLI * isqx.S]]
DurationS: TypeAlias = Annotated[_T, isqx.DURATION(isqx.S)]
UtcOffsetS: TypeAlias = Annotated[_T, isqx.DURATION["utc_offset"](isqx.S)]
DistanceM: TypeAlias = Annotated[_T, isqx.DISTANCE(isqx.M)]

LatitudeDeg: TypeAlias = Annotated[_T, isqx.LATITUDE(isqx.DEG)]
LongitudeDeg: TypeAlias = Annotated[_T, isqx.LONGITUDE(isqx.DEG)]
BearingDeg: TypeAlias = Annotated[_T, isqx.aerospace.HEADING(isqx.DEG)]
GroundTrackDeg: TypeAlias = Annotated[_T, isqx.aerospace.GROUND_TRACK(isqx.DEG)]
BankAngleDeg: TypeAlias = Annotated[_T, isqx.aerospace.BANK_ANGLE(isqx.DEG)]

AltitudeM: TypeAlias = Annotated[_T, isqx.aerospace.PRESSURE_ALTITUDE(isqx.M)]
AltitudeFt: TypeAlias = Annotated[
    _T, isqx.aerospace.PRESSURE_ALTITUDE(isqx.usc.FT)
]
GeometricAltitudeM: TypeAlias = Annotated[
    _T, isqx.aerospace.GEOMETRIC_ALTITUDE(isqx.M)
]
PressureAltimeterHpa: TypeAlias = Annotated[
    _T, isqx.aerospace.PRESSURE_ALTIMETER(isqx.HECTO * isqx.PA)
]
StaticTemperatureC: TypeAlias = Annotated[
    _T, isqx.aerospace.STATIC_TEMPERATURE(isqx.CELSIUS)
]

GroundSpeedKmh: TypeAlias = Annotated[
    _T,
    isqx.aerospace.GROUND_SPEED(isqx.KILO * isqx.M * isqx.HOUR**-1),
]
GroundSpeedKt: TypeAlias = Annotated[
    _T, isqx.aerospace.GROUND_SPEED(isqx.usc.KNOT)
]
GroundSpeedMph: TypeAlias = Annotated[
    _T, isqx.aerospace.GROUND_SPEED(isqx.usc.MPH)
]
IndicatedAirSpeedKt: TypeAlias = Annotated[
    _T, isqx.aerospace.INDICATED_AIRSPEED(isqx.usc.KNOT)
]
TrueAirSpeedKt: TypeAlias = Annotated[
    _T, isqx.aerospace.TRUE_AIRSPEED(isqx.usc.KNOT)
]
WindSpeedKt: TypeAlias = Annotated[_T, isqx.aerospace.WIND_SPEED(isqx.usc.KNOT)]
# TODO: does fr24 return inertial/GNSS or barometric?
VerticalSpeedFpm: TypeAlias = Annotated[
    _T, isqx.aerospace.VERTICAL_RATE(isqx.usc.FT * isqx.MIN**-1)
]
VerticalSpeedMps: TypeAlias = Annotated[
    _T, isqx.aerospace.VERTICAL_RATE(isqx.M * isqx.S**-1)
]

MachTimes1K: TypeAlias = Annotated[_T, isqx.MACH_NUMBER * 1000]



import httpx
import pytest

from fr24.grpc import (
    nearest_flights,
)
from fr24.proto import parse_data
from fr24.proto.headers import get_grpc_headers
from fr24.proto.v1_pb2 import (
    Geolocation,
    NearestFlightsRequest,
    NearestFlightsResponse,
)

HEADERS = httpx.Headers(get_grpc_headers(auth=None))


# NOTE: this fixture already exists in `tests/api/conftest.py`
# TODO: remove once services are implemented
@pytest.fixture
async def nearest_flights_response(
    client: httpx.AsyncClient,
) -> NearestFlightsResponse:
    message = NearestFlightsRequest(
        location=Geolocation(lat=22.31257, lon=113.92708),
        radius=10000,
        limit=1500,
    )
    response = await nearest_flights(client, message, HEADERS)
    return parse_data(response.content, NearestFlightsResponse).unwrap()


@pytest.mark.skip(
    reason="Server began to return empty `DATA` frames since Sep 2024"
)
@pytest.mark.anyio
async def test_live_trail(
    nearest_flights_response: NearestFlightsResponse, client: httpx.AsyncClient
) -> None:
    from fr24.grpc import live_trail
    from fr24.proto.v1_pb2 import LiveTrailRequest, LiveTrailResponse

    flight_id = nearest_flights_response.flights_list[0].flight.flightid
    message = LiveTrailRequest(flight_id=flight_id)
    response = await live_trail(client, message, HEADERS)
    result = parse_data(response.content, LiveTrailResponse)
    data = result.unwrap()
    assert len(data.radar_records_list)


@pytest.mark.skip(reason="Private API, does not return data.")
@pytest.mark.anyio
async def test_search_index(client: httpx.AsyncClient) -> None:
    from fr24.grpc import search_index
    from fr24.proto.v1_pb2 import (
        FetchSearchIndexRequest,
        FetchSearchIndexResponse,
    )

    message = FetchSearchIndexRequest()
    response = await search_index(client, message, HEADERS)
    result = parse_data(response.content, FetchSearchIndexResponse)
    assert result.is_ok()  # fails, empty data frame
    data = result.unwrap()
    assert data


@pytest.mark.skip(reason="Private API, does not return data.")
@pytest.mark.anyio
async def test_historic_trail(
    nearest_flights_response: NearestFlightsResponse, client: httpx.AsyncClient
) -> None:
    from fr24.grpc import historic_trail
    from fr24.proto.v1_pb2 import HistoricTrailRequest, HistoricTrailResponse

    flight_id = nearest_flights_response.flights_list[0].flight.flightid
    message = HistoricTrailRequest(flight_id=flight_id)
    response = await historic_trail(client, message, HEADERS)
    result = parse_data(response.content, HistoricTrailResponse)

    assert result.is_ok()  # read timeout
    data = result.unwrap()
    from fr24.proto.v1_pb2 import HistoricTrailResponse

    assert isinstance(data, HistoricTrailResponse)


def test_parse_data_grpc_status_error() -> None:
    """Fix for: https://github.com/abc8747/fr24/issues/68"""
    from fr24.proto import GrpcError, parse_data
    from fr24.proto.v1_pb2 import FollowFlightResponse

    error_data = (
        b"\x80\x00\x00\x00/grpc-status:5\r\ngrpc-message:Flight not found!\r\n"
    )

    result = parse_data(error_data, FollowFlightResponse)

    assert result.is_err()
    err = result.err()
    assert isinstance(err, GrpcError)
    assert err.status == 5
    assert err.status_message == b"Flight not found!"



from __future__ import annotations

import logging
from dataclasses import replace
from typing import TYPE_CHECKING

import httpx

from .authentication import login
from .cache import PATH_CACHE, FR24Cache
from .configuration import FP_CONFIG_FILE, PATH_CONFIG
from .grpc import BoundingBox
from .json import get_json_headers
from .proto.headers import get_grpc_headers
from .service import ServiceFactory
from .static.bbox import LNGS_WORLD_STATIC
from .types.json import Authentication
from .utils import dataclass_frozen

if TYPE_CHECKING:
    from typing import Any, Literal

    from typing_extensions import Self

    from .types.json import (
        TokenSubscriptionKey,
        UsernamePassword,
    )

logger = logging.getLogger(__name__)


class FR24:
    def __init__(
        self,
        client: httpx.AsyncClient | None = None,
    ) -> None:
        """See docs [quickstart](../usage/quickstart.md#initialisation).

        :param client: The `httpx` client to use. If not provided, a
            new one will be created with HTTP/2 enabled by default. It is
            highly recommended to use `http2=True` to avoid
            [464 errors](https://github.com/abc8747/fr24/issues/23#issuecomment-2125624974)
            and to be consistent with the browser.
        """
        auth = None
        self.http = HTTPClient(
            httpx.AsyncClient(http2=True) if client is None else client,
            auth=auth,
            grpc_headers=httpx.Headers(get_grpc_headers(auth=auth)),
            json_headers=httpx.Headers(get_json_headers()),
        )
        """The HTTP client for use in requests"""
        self._build_factory(self.http)

    def _build_factory(self, http: HTTPClient) -> None:
        factory = ServiceFactory(http)
        self.flight_list = factory.build_flight_list()
        """Flight list service."""
        self.playback = factory.build_playback()
        """Playback service."""
        self.live_feed = factory.build_live_feed()
        """Live feed service."""
        self.live_feed_playback = factory.build_live_feed_playback()
        """Live feed playback service."""
        self.airport_list = factory.build_airport_list()
        """Airport list service."""
        self.find = factory.build_find()
        """Find service."""
        self.nearest_flights = factory.build_nearest_flights()
        """Nearest flights service."""
        self.live_flights_status = factory.build_live_flights_status()
        """Live flights status service."""
        self.follow_flight = factory.build_follow_flight()
        """Follow flight service."""
        self.top_flights = factory.build_top_flights()
        """Top flights service."""
        self.flight_details = factory.build_flight_details()
        """Flight details service."""
        self.playback_flight = factory.build_playback_flight()
        """Playback flight service."""

    async def login(
        self,
        creds: (
            TokenSubscriptionKey | UsernamePassword | None | Literal["from_env"]
        ) = "from_env",
    ) -> None:
        """:param creds: Reads credentials from the environment variables or the
        config file if `creds` is set to `"from_env"` (default). Otherwise,
        provide the credentials directly.
        """
        self.http = await self.http.with_login(creds)
        self._build_factory(self.http)
        if (auth := self.http.auth) is None:
            logger.info("no authentication provided, using anonymous access")
        elif "subscription_key" in (message := auth.get("message", "")):
            logger.info(message)
        else:
            logger.info("using username and password")

    async def __aenter__(self) -> Self:
        await self.http.__aenter__()
        return self

    async def __aexit__(self, *args: Any) -> None:
        await self.http.__aexit__(*args)


@dataclass_frozen
class HTTPClient:
    """An HTTPX client for making requests to the API."""

    client: httpx.AsyncClient
    auth: Authentication | None
    grpc_headers: httpx.Headers
    json_headers: httpx.Headers

    async def with_login(
        self,
        creds: (
            TokenSubscriptionKey | UsernamePassword | Literal["from_env"] | None
        ) = "from_env",
    ) -> HTTPClient:
        auth = await login(self.client, creds)
        return replace(
            self,
            auth=auth,
            grpc_headers=httpx.Headers(get_grpc_headers(auth=auth)),
            json_headers=httpx.Headers(get_json_headers()),
        )

    async def __aenter__(self) -> HTTPClient:
        return self

    async def __aexit__(self, *args: Any) -> None:
        if self.client is not None:
            await self.client.aclose()


BBOX_FRANCE_UIR = BoundingBox(42.0, 52.0, -8.0, 10.0)
"""Bounding box for france UIR"""

BBOXES_WORLD_STATIC = [
    BoundingBox(-90, 90, LNGS_WORLD_STATIC[i], LNGS_WORLD_STATIC[i + 1])
    for i in range(len(LNGS_WORLD_STATIC) - 1)
]
"""Default static bounding boxes covering the entire world"""

__all__ = [
    "BBOXES_WORLD_STATIC",
    "BBOX_FRANCE_UIR",
    "FP_CONFIG_FILE",
    "FR24",
    "PATH_CACHE",
    "PATH_CONFIG",
    "BoundingBox",
    "FR24Cache",
    "HTTPClient",
]



import httpx

from ..types.static import AircraftFamily, Airlines, Airports, Countries
from ..utils import DEFAULT_HEADERS

DEFAULT_HEADERS_STATIC = {
    **DEFAULT_HEADERS,
    "Accept": "*/*",
}


async def fetch_aircraft_family(client: httpx.AsyncClient) -> AircraftFamily:
    request = httpx.Request(
        "GET",
        "https://www.flightradar24.com/mobile/aircraft-family",
        headers=DEFAULT_HEADERS_STATIC,
    )
    data = await client.send(request)
    data.raise_for_status()
    return data.json()  # type: ignore[no-any-return]


async def fetch_airlines(client: httpx.AsyncClient) -> Airlines:
    request = httpx.Request(
        "GET",
        "https://www.flightradar24.com/mobile/airlines",
        headers=DEFAULT_HEADERS_STATIC,
    )
    data = await client.send(request)
    data.raise_for_status()
    return data.json()  # type: ignore[no-any-return]


async def fetch_airports(
    client: httpx.AsyncClient, major_version: int = 4, minor_version: int = 0
) -> Airports:
    request = httpx.Request(
        "GET",
        f"https://www.flightradar24.com/mobile/airports/format/{major_version}",
        params={"version": minor_version},
        headers=DEFAULT_HEADERS_STATIC,
    )
    data = await client.send(request)
    data.raise_for_status()
    return data.json()  # type: ignore[no-any-return]


async def fetch_countries(client: httpx.AsyncClient) -> Countries:
    request = httpx.Request(
        "GET",
        "https://www.flightradar24.com/mobile/countries",
        headers=DEFAULT_HEADERS_STATIC,
    )
    data = await client.send(request)
    data.raise_for_status()
    return data.json()  # type: ignore[no-any-return]


# NOTE: previously, the code downloaded the static json data directly into
# `Path(__file__).parent`. This gives a convenient way to access the data,
# but is often outdated. To avoid committing large diffs, we should instead
# save them into the user cache instead.



# ruff: noqa
# mypy: disable-error-code="top-level-await, no-redef"
# %%
# --8<-- [start:script0]
import httpx
from fr24.grpc import LiveFeedParams, BoundingBox, live_feed
from fr24.proto.v1_pb2 import LiveFeedResponse
from fr24.proto import parse_data
from fr24.proto.headers import get_grpc_headers


async def france_data() -> LiveFeedResponse:
    headers = httpx.Headers(get_grpc_headers(auth=None))
    async with httpx.AsyncClient() as client:
        params = LiveFeedParams(
            bounding_box=BoundingBox(north=50, west=-7, south=40, east=10)
        )
        response = await live_feed(client, params, headers)
        result = parse_data(response.content, LiveFeedResponse)
        return result.unwrap()  # (1)!


data = await france_data()
data
# --8<-- [end:script0]
# %%
"""
# --8<-- [start:output0]
flights_list {
  flightid: 941292228
  lat: 40.0058289
  lon: -5.02393532
  track: 225
  alt: 26800
  speed: 427
  icon: A320
  timestamp: 1732481230
  callsign: "IBS15MA"
  extra_info {
    flight: "IB1589"
    reg: "EC-OAS"
    route {
      from: "MAD"
      to: "TFN"
    }
    type: "A21N"
  }
  position_buffer {
    recent_positions_list {
      delta_lat: -418
      delta_lon: -562
      delta_ms: 3000
    }
...
      delta_lon: -1958
      delta_ms: 10000
    }
  }
}
# --8<-- [end:output0]
"""
# %% [markdown]
# # explore in JSON format

# %%
# --8<-- [start:script1]
from google.protobuf.json_format import MessageToDict

MessageToDict(data)["flightsList"]
# --8<-- [end:script1]
# %%
"""
# --8<-- [start:output1]
[{'flightid': 941292228,
  'lat': 40.00583,
  'lon': -5.0239353,
  'track': 225,
  'alt': 26800,
  'speed': 427,
  'icon': 'A320',
  'timestamp': 1732481230,
  'callsign': 'IBS15MA',
  'extraInfo': {'flight': 'IB1589',
   'reg': 'EC-OAS',
   'route': {'from': 'MAD', 'to': 'TFN'},
   'type': 'A21N'},
  'positionBuffer': {'recentPositionsList': [{'deltaLat': -418,
     'deltaLon': -562,
     'deltaMs': 3000}]}},
 {'flightid': 941277503,
  'lat': 40.050613,
  'lon': -6.7190137,
  'track': 24,
  'alt': 38000,
  'speed': 481,
  'icon': 'A320',
  'timestamp': 1732481230,
  'callsign': 'VOE2RY',
...
  'positionBuffer': {'recentPositionsList': [{'deltaLat': 315,
     'deltaLon': -660,
     'deltaMs': 3000},
    {'deltaLat': 553, 'deltaLon': -1170, 'deltaMs': 6000},
    {'deltaLat': 929, 'deltaLon': -1958, 'deltaMs': 10000}]}}]
# --8<-- [end:output1]
"""
# %%



from __future__ import annotations

import base64
import configparser
import json
import logging
import os
import time
from datetime import datetime, timezone
from pathlib import Path
from typing import Literal

import httpx

from .configuration import FP_CONFIG_FILE
from .types.json import (
    Authentication,
    TokenSubscriptionKey,
    UsernamePassword,
)
from .utils import DEFAULT_HEADERS

logger = logging.getLogger(__name__)


def get_credentials(
    fp_config_file: Path = FP_CONFIG_FILE,
) -> TokenSubscriptionKey | UsernamePassword | None:
    """Reads credentials from the environment variables, overriding it with
    the config file if it exists.
    """
    username = os.environ.get("fr24_username", None)
    password = os.environ.get("fr24_password", None)
    subscription_key = os.environ.get("fr24_subscription_key", None)
    token = os.environ.get("fr24_token", None)

    if fp_config_file.exists():
        config = configparser.ConfigParser()
        config.read(fp_config_file.as_posix())

        username = config.get("global", "username", fallback=None)
        password = config.get("global", "password", fallback=None)
        subscription_key = config.get(
            "global", "subscription_key", fallback=None
        )
        token = config.get("global", "token", fallback=None)

    if username and password:
        return {"username": username, "password": password}
    if subscription_key and token:
        return {"subscriptionKey": subscription_key, "token": token}
    return None


async def login(
    client: httpx.AsyncClient,
    creds: (
        TokenSubscriptionKey | UsernamePassword | None | Literal["from_env"]
    ) = "from_env",
    fp_config_file: Path = FP_CONFIG_FILE,
) -> None | Authentication:
    """Read credentials and returns the credentials needed to access the API.

    By default, credentials are read from the environment variables or the
    config file if `creds_override` is not set. Then, if the credentials:
    - `username` and `password` is set: makes a POST request to the login
    endpoint
    - `subscription_key` and `token` is set: returns immediately
    - otherwise, `None` is returned
    """
    creds = get_credentials(fp_config_file) if creds == "from_env" else creds

    if creds is None:
        return None
    if (u := creds.get("username")) and (p := creds.get("password")):
        return await login_with_username_password(client, u, p)  # type: ignore[arg-type]
    if s := creds.get("subscriptionKey"):
        t = creds.get("token")
        return await login_with_token_subscription_key(client, s, t)  # type: ignore[arg-type]

    logger.warning(
        "expected username+password or subscriptionKey+Optional[token] pair,"
        "but one or both are missing. falling back to anonymous access."
    )
    return None


async def login_with_username_password(
    client: httpx.AsyncClient,
    username: str,
    password: str,
) -> Authentication:
    """Retrieve bearer token and subscription key from the API.

    Bearer: `json['userData']['accessToken']`
    `token=` query param: `json['userData']['subscriptionKey']`
    """
    response = await client.post(
        "https://www.flightradar24.com/user/login",
        data={"email": username, "password": password},
        headers=DEFAULT_HEADERS,
    )
    response.raise_for_status()
    return response.json()  # type: ignore


async def login_with_token_subscription_key(
    _client: httpx.AsyncClient,
    subscription_key: str,
    token: str | None,
) -> Authentication | None:
    """Login with subscription key and/or token.
    Falls back to anonymous access if token is expired or invalid.
    """
    if token is None:
        return {
            "userData": {
                "subscriptionKey": subscription_key,
            },
            "message": "using environment `subscription_key`",
        }

    try:
        payload = json.loads(base64.b64decode(token.split(".")[1]))
    except Exception as e:
        logger.warning(
            f"failed to parse token: {e}. falling back to anonymous access"
        )
        return None

    exp = payload["exp"]
    exp_f = datetime.fromtimestamp(exp, timezone.utc).isoformat()
    if time.time() > exp:
        logger.warning(
            f"token has expired at {exp_f}. falling back to anonymous access"
        )
        return None

    return {
        "user": {
            "id": payload.get("userId"),
        },
        "userData": {
            "subscriptionKey": subscription_key,
            "accessToken": token,
            "dateExpires": exp,
        },
        "message": "using environment `subscription_key` and `token`",
    }



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: fr24/proto/_nearest_flights.proto
# Protobuf Python Version: 6.33.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    33,
    0,
    '',
    'fr24/proto/_nearest_flights.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from fr24.proto import _common_pb2 as fr24_dot_proto_dot___common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n!fr24/proto/_nearest_flights.proto\x1a\x18\x66r24/proto/_common.proto\"\'\n\x0bGeolocation\x12\x0b\n\x03lat\x18\x01 \x01(\x02\x12\x0b\n\x03lon\x18\x02 \x01(\x02\"V\n\x15NearestFlightsRequest\x12\x1e\n\x08location\x18\x01 \x01(\x0b\x32\x0c.Geolocation\x12\x0e\n\x06radius\x18\x02 \x01(\r\x12\r\n\x05limit\x18\x03 \x01(\r\"A\n\x0cNearbyFlight\x12\x1f\n\x06\x66light\x18\x01 \x01(\x0b\x32\x0f._common.Flight\x12\x10\n\x08\x64istance\x18\x02 \x01(\r\"=\n\x16NearestFlightsResponse\x12#\n\x0c\x66lights_list\x18\x01 \x03(\x0b\x32\r.NearbyFlightb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'fr24.proto._nearest_flights_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_GEOLOCATION']._serialized_start=63
  _globals['_GEOLOCATION']._serialized_end=102
  _globals['_NEARESTFLIGHTSREQUEST']._serialized_start=104
  _globals['_NEARESTFLIGHTSREQUEST']._serialized_end=190
  _globals['_NEARBYFLIGHT']._serialized_start=192
  _globals['_NEARBYFLIGHT']._serialized_end=257
  _globals['_NEARESTFLIGHTSRESPONSE']._serialized_start=259
  _globals['_NEARESTFLIGHTSRESPONSE']._serialized_end=320
# @@protoc_insertion_point(module_scope)
